private import mymath;
private import myio;
private import texture;
private import log;
private import rc;
private import camera;
private import fpstimer;
private import letter;
private import vector;
private import SDL;
private import SDL_pad;
private import opengl;
private import openglu;
//private import glee;
private import std.string;
private import std.math;

private import readygo;

Texture tex;

Main_ Main;

class Main_ {
    static this() {
        Main = new Main_();
    }

    int screenWidth_, screenHeight_;

    FpsTimer timer;

    Rc rc;

    ReadyGo[] readygos;

    int score;

	void initSDL() {
		Log.info("init SDL");

		int ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE);
		if (ret != 0) Log.errorSDL("cannot init SDL");

		ret = SDL_InitSubSystem(SDL_INIT_JOYSTICK);
		if (ret != 0) {
			Log.warnSDL("cannot init joystick");
		}

/*
  useAudio_ = true;
  ret = SDL_InitSubSystem(SDL_INIT_AUDIO);
  if (ret != 0) {
  Log.warnSDL("cannot init audio");
  useAudio_ = false;
  }
*/

		SDL_VideoInfo* info = SDL_GetVideoInfo();

        int bpp = rc.getInt("screenBpp");
        char[] bppMsg;

		if (bpp == -1) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using the \"best\" bpp: " ~ tostr(bpp));
		}
		else if (bpp == 0) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using current bpp");
		}
		else {
			Log.info("using user set bpp: " ~ tostr(bpp));
		}

		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
		SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

        if (rc.getInt("motionBlurEnable")) {
            SDL_GL_SetAttribute( SDL_GL_ACCUM_RED_SIZE, 5 );
            SDL_GL_SetAttribute( SDL_GL_ACCUM_GREEN_SIZE, 5 );
            SDL_GL_SetAttribute( SDL_GL_ACCUM_BLUE_SIZE, 5 );
        }

		screenWidth_ = rc.getInt("screenWidth");
		screenHeight_ = rc.getInt("screenHeight");
		int screenFull = rc.getInt("screenFull");

		int flags = SDL_OPENGL;

		if (screenFull) {
			flags |= SDL_FULLSCREEN;
		}

        Log.info("set video mode " ~
                 "size = " ~ tostr(screenWidth_) ~
                 "x" ~ tostr(screenHeight_) ~
                 ", bpp = " ~ tostr(bpp) ~
                 ", flags = " ~ tostr(flags));

		if (SDL_SetVideoMode(screenWidth_, screenHeight_,
							 bpp, flags) == null)
		{
			Log.errorSDL("cannot create screen");
		}

		info = SDL_GetVideoInfo();
		Log.info("set bpp: " ~ tostr(cast(int)info.vfmt.BitsPerPixel));

        SDL_WM_SetCaption(`pbg`, null);

        Log.info("pad init");
        Pad_Init();

        Pad_LoadDefault("padrc");
	}

	void initGL() {
        float bgr = rc.getFloat("clearColR");
        float bgg = rc.getFloat("clearColG");
        float bgb = rc.getFloat("clearColB");
        if (rc.getInt("reverseColor")) {
            glClearColor(1-bgr, 1-bgg, 1-bgb, 0);

            //glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
            //glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        }
        else {
            glClearColor(bgr, bgg, bgb, 0);
        }

		glViewport(0, 0, screenWidth_, screenHeight_);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		float asp = cast(float)screenWidth_ / screenHeight_;
		gluPerspective(45.0, asp, 0.1, 1000);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
//		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//		glBlendFunc(GL_ONE, GL_ZERO);

//        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
//        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

		glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

		// from PARSEC47
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH);
		glDisable(GL_DEPTH_TEST);
        glDisable(GL_TEXTURE_2D);

        glLineWidth(rc.getInt("bgLineWidth"));
	}

    int main(char[][] args) {
        Log.init();

        rc = new Rc();
        rc.addResourceFile(`pbg.rc`);
        initSDL();
        initGL();
        initLetter();

        int enableSystemInfo = rc.getInt("enableSystemInfo");

        int cameraId = rc.getInt("cameraId");
        Camera camera;
        if (cameraId == -1) {
            camera = new Camera();
        }
        else {
            camera = new Camera(cameraId);
        }
        timer = new FpsTimer(true);
		int frameSkip = rc.getInt("frameSkip");
        if (args.length >= 2) frameSkip = 0;
        if (frameSkip) {
            int fps = rc.getInt("fps");
            timer.setFps(fps);
        }
        else {
            timer.setFrameSkip(false);
        }

        Texture.init();

        int motionBlurEnable = rc.getInt("motionBlurEnable");
        int motionBlurFrames = rc.getInt("motionBlurFrames");
        float motionBlurAccum = rc.getFloat("motionBlurAccum");
        float motionBlurMult = rc.getFloat("motionBlurMult");
        if (motionBlurEnable && frameSkip) {
            Log.warn("motion blur enabled. so force disables frame skip.");
            frameSkip = false;
            timer.setFrameSkip(false);
        }

        Uint32 startTicks = SDL_GetTicks();

        int turn;
        score = -1;
        for (turn = 0; turn < 10000000; turn++) {
            Pad_Pump(PAD_PUMP_EVERYTHING);
            if (Pad_GetButtonPushed(PAD_END_BUTTON)) {
                break;
            }
            if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
                if (score == -1) {
                    int size = screenWidth_*screenHeight_*3;
                    ubyte[] pixels = new ubyte[size];
                    glReadPixels(0, 0, screenWidth_, screenHeight_,
                                 GL_RGB, GL_UNSIGNED_BYTE, &pixels[0]);
                    int sum = 0;
                    for (int i = 0; i < size; i++) {
                        sum += pixels[i];
                    }

                    score = sum/size*10000/256;
                }
                else {
                    score = -1;
                    readygos = null;
                }
            }
/*
            if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
                camera.end();
            }
            if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
                if (camera.stop) Log.info("camera move");
                else Log.info("camera stop");
                camera.stop = !camera.stop;
            }
*/

            if (score == -1) {
                if (rnd(10) == 0) {
                    float rp() { return rnd(300.0) - 150.0; }
                    float rs() { return rnd(4.0) - 2.0; }
                    ReadyGo readygo = new ReadyGo(Vec(rp(), rp(), rs()),
                                                  Vec(rs(), rs(), rs()));
                    readygos ~= readygo;
                }

                ReadyGo[] nextR;
                foreach (ReadyGo readygo; readygos) {
                    readygo.proc();
                    if (readygo.alive) nextR ~= readygo;
                }
                readygos = nextR;
            }

            if (!timer.shouldSkip()) {
                glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
                glMatrixMode(GL_MODELVIEW);
                glLoadIdentity();

                if (turn < 200 && cameraId == -1) {
                    gluLookAt(240, 240, -500, 240, 240, 0, 0, -1, 0);
                }
                else {
                    camera.set();
                    glTranslatef(80, 0, 0);
                }

                foreach (ReadyGo readygo; readygos)
                    readygo.draw((score == -1) ? 3.0 : 0.3);

                if (enableSystemInfo) drawSystemInfo();

                // drawBox();

                if (motionBlurEnable) {
                    glAccum(GL_ACCUM, motionBlurAccum);
                    int t = turn % motionBlurFrames;
                    if (t == motionBlurFrames-1) {
                        glAccum(GL_RETURN, 1.0);
                        SDL_GL_SwapBuffers();
                        glClear(GL_ACCUM_BUFFER_BIT);
                    }
                    else {
                        glAccum(GL_MULT, motionBlurMult);
                    }
                }
                else {
                    SDL_GL_SwapBuffers();
                }
            }
            timer.wait();
//            if (frameSkip) timer.wait();
//        SDL_Delay(16);
        }

        Uint32 endTicks = SDL_GetTicks();
        printf("%d FPS.\n", turn*1000/(endTicks-startTicks));

        SDL_Quit();

        return 0;
    }

    void drawSystemInfo() {
        glLoadIdentity();
        gluLookAt(240, 240, -500, 240, 240, 0, 0, -1, 0);

        int x = -35;
        int y = 45;
        glDisable(GL_TEXTURE_2D);
        drawString(`FPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getTrueFps(), x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`FRAMES`, x,y, 10,12, 1,1,1);
        drawNum(timer.getTurn(), x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`DROPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getDroppedTurn(), x+65,y, 10,12, 1,1,1);
/*
        y += 15;
        drawString(`SDROPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getSystemDroppedTurn(), x+65,y, 10,12, 1,1,1);
*/
        int cnt = readygos.length;
        y += 15;
        drawString(`OBJS`, x,y, 10,12, 1,1,1);
        drawNum(cnt, x+65,y, 10,12, 1,1,1);

        if (score != -1) {
            drawNum(score, 0, screenHeight_/2, 60, 100, 1,1,1);
            drawString(`PTS`, 260, screenHeight_/2, 60, 100, 1,1,1);
        }
    }
}

int main(char[][] args) {
    return Main.main(args);
}

